import re
from bs4 import BeautifulSoup, Tag, NavigableString
from difflib import SequenceMatcher

REPLY_HEADER_REGEX = re.compile(
    r'((On .+?wrote:)|([0-9]{1,2}:[0-9]{2} ?[ap]m,? .+? <.+?>:))',
    re.IGNORECASE | re.DOTALL
)

def normalize_text(text):
    # Lowercase, strip, collapse whitespace, remove non-breaking spaces and invisible chars
    text = text.replace('\xa0', ' ').replace('\u200b', ' ')
    text = re.sub(r'\s+', ' ', text)
    return text.lower().strip()

def is_similar(a, b, threshold=0.85):
    return SequenceMatcher(None, a, b).ratio() > threshold

def get_direct_text(element):
    # Clone the element, remove all nested blockquotes, then extract all text
    html = str(element)
    print('---BLOCKQUOTE HTML---')
    print(html[:1000])
    print('---BLOCKQUOTE HTML BYTES---')
    print(repr(html.encode(errors='replace')[:1000]))
    try:
        clone = BeautifulSoup(html, 'lxml')
    except Exception:
        clone = BeautifulSoup(html, 'html.parser')
    for nested in clone.find_all('blockquote'):
        nested.decompose()
    raw_text = clone.get_text()
    if not raw_text.strip():
        print('Fallback: using element.get_text()')
        raw_text = element.get_text()
    if not raw_text.strip():
        print('Fallback: using regex extraction')
        m = re.search(r'<blockquote.*?>(.*?)</blockquote>', html, re.DOTALL | re.IGNORECASE)
        if m:
            raw_text = BeautifulSoup(m.group(1), 'lxml').get_text()
    print('---BLOCKQUOTE RAW TEXT---')
    print(repr(raw_text[:1000]))
    text = raw_text.strip()
    print('---BLOCKQUOTE EXTRACTED TEXT---')
    print(text[:1000])
    return text

def remove_reply_header(block):
    # Remove previous siblings that are reply headers or <br> tags
    prev = block.previous_sibling
    while prev:
        # If it's a <br> or whitespace, remove and continue
        if isinstance(prev, Tag) and prev.name == 'br':
            to_remove = prev
        elif isinstance(prev, Tag) and prev.name in ('div', 'span'):
            # Check text content for reply header
            if REPLY_HEADER_REGEX.search(prev.get_text()):
                to_remove = prev
            else:
                break
        elif isinstance(prev, NavigableString):
            if REPLY_HEADER_REGEX.search(str(prev)) or str(prev).strip() == '':
                to_remove = prev
            else:
                break
        else:
            break
        next_prev = prev.previous_sibling
        prev.extract()
        prev = next_prev

def filter_redundant_quoted_content(html: str, chat, references: list[str]):
    """
    Remove quoted blocks from html if their direct text matches referenced messages in the chat.
    Only removes a quote if it matches the referenced message (by id in references).
    Handles nested blockquotes recursively.
    """
    if not html or not references or not chat:
        return html
    from unicom.models import Message
    soup = BeautifulSoup(html, 'html.parser')

    def process_blockquote(block, ref_ids):
        if not ref_ids:
            return
        ref_id = ref_ids[0]
        try:
            msg = chat.messages.filter(id=ref_id).first()
        except Exception:
            msg = None
        if not msg:
            return
        block_direct_text = normalize_text(get_direct_text(block))
        msg_html = msg.html or msg.text or ''
        msg_text = normalize_text(BeautifulSoup(msg_html, 'html.parser').get_text())
        print('---BLOCKQUOTE---')
        print(block_direct_text[:500])
        print('---MSG TEXT---')
        print(msg_text[:500])
        sim = SequenceMatcher(None, block_direct_text, msg_text).ratio()
        print(f'Similarity: {sim}')
        if sim > 0.85:
            remove_reply_header(block)
            block.decompose()
            return True
        return False

    def recursive_filter(element, ref_ids):
        for bq in element.find_all('blockquote', recursive=False):
            if ref_ids:
                matched = process_blockquote(bq, ref_ids)
                if matched:
                    ref_ids.pop(0)
                    continue
                recursive_filter(bq, ref_ids[1:])

    recursive_filter(soup, list(references))
    return str(soup) 


# import time
# import logging
# from imapclient import IMAPClient, SEEN
# from django.core.management.base import BaseCommand
# from django.conf import settings

# from robopower.services.email.save_email_message import save_email_message

# logger = logging.getLogger(__name__)

# class Command(BaseCommand):
#     help = "Listen to an IMAP account via IDLE and hand off new emails."

#     def handle(self, *args, **options):
#         while True:
#             try:
#                 with IMAPClient(settings.IMAP_HOST, port=settings.IMAP_PORT, ssl=settings.IMAP_USE_SSL) as server:
#                     print(f"Logging in to {settings.IMAP_HOST}:{settings.IMAP_PORT} as {settings.IMAP_USER} with pwd {settings.IMAP_PASSWORD}")
#                     server.login(settings.IMAP_USER, settings.IMAP_PASSWORD)
#                     server.select_folder('INBOX')
#                     self.stdout.write("✅ Connected to IMAP, entering IDLE…")

#                     while True:
#                         try:
#                             server.idle()                             # enter IDLE
#                             responses = server.idle_check(timeout=300)  # wake up at least every 5m
#                         except (ConnectionResetError, OSError) as e:
#                             logger.warning("IMAP idle connection lost: %s", e)
#                             break  # break inner loop to reconnect
#                         finally:
#                             # if idle() succeeded we need to terminate it
#                             # (noop if connection already gone)
#                             try:
#                                 server.idle_done()
#                             except Exception:
#                                 pass

#                         if not responses:
#                             continue

#                         # process all new unseen messages
#                         uids = server.search(['UNSEEN'])
#                         for uid in uids:
#                             try:
#                                 resp = server.fetch(uid, ['BODY.PEEK[]'])
#                                 raw = resp[uid][b'BODY[]']
#                                 msg = save_email_message(raw)
#                                 self.stdout.write(f"💾 Saved email {msg.id} (uid={uid})")
#                                 server.add_flags(uid, [SEEN])
#                             except Exception as exc:
#                                 logger.exception("Failed to process UID %s: %s", uid, exc)

#             except Exception as e:
#                 logger.exception("Fatal IMAP error, reconnecting in 30s…")
#                 time.sleep(30)



from django.db import models
from django.db.models.fields.files import FileField, FieldFile
import hashlib

class DedupFieldFile(FieldFile):
    def save(self, name, content, save=True):
        # Compute hash
        content.seek(0)
        file_data = content.read()
        content.seek(0)
        file_hash = hashlib.sha256(file_data).hexdigest()
        # Set hash on instance
        setattr(self.instance, self.field.hash_field, file_hash)
        # Check for duplicate in same model
        Model = type(self.instance)
        existing = Model.objects.filter(**{self.field.hash_field: file_hash}).exclude(pk=self.instance.pk).first()
        if existing:
            # Point to existing file path, do NOT call super().save()
            self.name = getattr(existing, self.field.name).name
            setattr(self.instance, self.field.attname, self.name)
            # Save the model instance to persist the file path and hash
            self.instance.save(update_fields=[self.field.name, self.field.hash_field])
        else:
            super().save(name, content, save=save)

class DedupFileField(FileField):
    attr_class = DedupFieldFile

    def __init__(self, *args, hash_field='hash', **kwargs):
        self.hash_field = hash_field
        super().__init__(*args, **kwargs)

    def contribute_to_class(self, cls, name, **kwargs):
        super().contribute_to_class(cls, name, **kwargs)
        # Document: model must have a nullable, indexed (not unique) hash field with name self.hash_field

# Utility for model delete methods

def only_delete_file_if_unused(instance, file_field_name, hash_field_name):
    """
    Call this in your model's delete method before deleting the file.
    Deletes the file only if no other objects of the same model share the same hash.
    """
    Model = type(instance)
    file_field = getattr(instance, file_field_name)
    file_hash = getattr(instance, hash_field_name)
    if file_field and file_hash:
        others = Model.objects.filter(**{hash_field_name: file_hash}).exclude(pk=instance.pk)
        if not others.exists():
            file_field.delete(save=False)
# Usage:
# 1. Add a nullable, indexed (not unique) hash field to your model (e.g. hash = models.CharField(max_length=64, blank=True, null=True, db_index=True))
# 2. Use DedupFileField in place of FileField, e.g. file = DedupFileField(upload_to=..., hash_field='hash')
# 3. In your model's delete method, call only_delete_file_if_unused(self, 'file', 'hash') before super().delete() 


import requests
from uuid import uuid4
from unicom.services.get_public_origin import get_public_origin


def set_telegram_webhook(bot_credentials, allowed_updates=None):
    """
    Set the Telegram bot webhook.
    """
    TelegramCredentials = bot_credentials.config
    TELEGRAM_API_TOKEN = TelegramCredentials["TELEGRAM_API_TOKEN"]
    if "TELEGRAM_SECRET_TOKEN" not in TelegramCredentials:
        TelegramCredentials["TELEGRAM_SECRET_TOKEN"] = uuid4().hex
        bot_credentials.config["TELEGRAM_SECRET_TOKEN"] = TelegramCredentials["TELEGRAM_SECRET_TOKEN"]
    TELEGRAM_SECRET_TOKEN = TelegramCredentials["TELEGRAM_SECRET_TOKEN"]
    origin = get_public_origin()
    if not origin:
        raise ValueError("DJANGO_PUBLIC_ORIGIN is not set")
    webhook_url = f"{origin}/unicom/telegram/{bot_credentials.id}"
    print(f"Updating Telegram Webhook to {webhook_url}")
    url = f"https://api.telegram.org/bot{TELEGRAM_API_TOKEN}/setWebhook"

    data = {
        "url": webhook_url,
        "secret_token": TELEGRAM_SECRET_TOKEN
    }

    if allowed_updates:
        data["allowed_updates"] = allowed_updates

    response = requests.post(url, data=data)
    if response.status_code == 200:
        bot_credentials.confirmed_webhook_url = webhook_url
    return response.json()  # Returns the API response to check if the webhook was set correctly



from django.apps import AppConfig


class UnicomConfig(AppConfig):
    name = 'unicom'

    def ready(self):
        from unicom.services.email.IMAP_thread_manager import imap_manager
        import unicom.signals
        imap_manager.start_all()



from django.db import models


class Member(models.Model):
    """
    CRM model for customers/users with specific access levels and temporary tokens.
    """
    name = models.CharField(max_length=255)
    email = models.EmailField(
        blank=True,
        null=True,
        db_index=True,
        help_text="Indexed field for member identification"
    )
    phone = models.CharField(
        max_length=50,
        blank=True,
        null=True,
        db_index=True,
        help_text="Indexed field for member identification"
    )
    misc_tokens = models.JSONField(
        default=dict,
        blank=True,
        null=True,
        help_text="Secure storage for temporary authentication/access tokens"
    )
    allowed_categories = models.ManyToManyField(
        'unicom.RequestCategory',
        blank=True,
        related_name='members_with_access',
        help_text="Categories this member has explicit access to"
    )
    user = models.OneToOneField(
        'auth.User',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        related_name='member'
    )
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        ordering = ['name']
        indexes = [
            # Compound index for when searching both fields
            models.Index(fields=['email', 'phone'], name='member_contact_idx'),
        ]

    def __str__(self):
        return self.name 


from django.http import HttpResponse, JsonResponse
from django.views.decorators.csrf import csrf_exempt
from unicom.models import Update
from unicom.services.whatsapp.save_whatsapp_message import save_whatsapp_message
from unicom.services.whatsapp.save_whatsapp_message_status import save_whatsapp_message_status
from django.db import transaction
import hashlib
import hmac
import json


# TODO: Re-enable signature verification and subscription mechanism
@csrf_exempt
def whatsapp_webhook(request):
    if request.method == 'POST':
        signature = request.headers.get('X-Hub-Signature-256', '').split('=')[1]
        # computed_signature = hmac.new(FACEBOOK_APP_SECRET.encode(), request.body, hashlib.sha256).hexdigest()
        # if not hmac.compare_digest(computed_signature, signature):
        #     print("Invalid Signature!")
        #     JsonResponse({'status': 'verification failed'}, status=403)
        data = request.body.decode('utf-8')
        data_dict = json.loads(data)
        if data_dict.get("object") != "whatsapp_business_account":
            JsonResponse({'status': 'Error 430'}, status=500)
        notification_objects = data_dict.get("entry")
        if len(notification_objects) != 1:
            JsonResponse({'status': 'Error 431'}, status=500)
        update = Update(platform='WhatsApp', id=f'WhatsApp.{signature}', payload=data_dict)
        update.save()
        
    
        for notification_object in notification_objects:
            changes = notification_object["changes"]
            for change in changes:
                if change["field"] == 'messages':
                    change_value = change["value"]
                    if "messages" in change_value:
                        with transaction.atomic():
                            msg = save_whatsapp_message(change_value)
                            update.message = msg
                            # Mark update as from blocked account if no message was saved
                            if msg is None:
                                update.from_blocked_account = True
                            update.save()
                    if "statuses" in change_value:
                        with transaction.atomic():
                            msg = save_whatsapp_message_status(change_value)
                            update.message = msg
                            update.save()

        return JsonResponse({'status': 'Update Received'}, status=200)
    # if request.method == 'GET' and request.GET.get('hub.mode', 'null') == 'subscribe':
    #     if WHATSAPP_VERIFY_TOKEN is not None and WHATSAPP_VERIFY_TOKEN == request.GET.get('hub.verify_token', 'null'):
    #         challenge = request.GET.get('hub.challenge', 'null')
    #         return HttpResponse(challenge)

    return HttpResponse('Invalid request method.')



from unicom.models import Message, AccountChat, Channel, Request
from django.db import transaction
from django.db.models.signals import post_save, pre_save, post_delete
from unicom.services.email.IMAP_thread_manager import imap_manager
from unicom.services.chat_summary import update_chat_summary
from django.db import transaction
from django.dispatch import receiver
import threading
from django.utils import timezone
import logging
from imapclient import IMAPClient, SEEN
import imaplib




@receiver(pre_save, sender=Channel)
def channel_pre_save(sender, instance, **kwargs):
    # Save old config for comparison in post_save
    if instance.pk:
        try:
            old = sender.objects.get(pk=instance.pk)
            instance._old_config = old.config
        except sender.DoesNotExist:
            instance._old_config = None


@receiver(post_save, sender=Channel)
def run_channel_after_insert(sender, instance, created, **kwargs):
    # Check if created or config changed
    def validate_and_then_start():
        instance.validate()
        imap_manager.restart(instance)
    config_changed = not created and getattr(instance, '_old_config', None) != instance.config
    if created or config_changed:
        transaction.on_commit(lambda: threading.Thread(target=validate_and_then_start).start())
 
@receiver(post_delete, sender=Channel)
def run_channel_after_delete(sender, instance, **kwargs):
    # Stop the IMAP listener thread for the channel
    imap_manager.stop(instance)
    print(f"Channel {instance.pk} deleted, IMAP listener stopped.")


@receiver(post_save, sender=Message)
def create_request_from_message(sender, instance, created, **kwargs):
    """
    Signal handler to create a Request object when a new Message is created.
    Only creates a Request for incoming messages (is_outgoing=False).
    """
    if not created:
        return

    # Update chat summary fields
    update_chat_summary(instance)
    
    # Skip request creation if it's an outgoing message
    if instance.is_outgoing:
        return

    # Extract contact information based on platform
    email = None
    phone = None
    
    if instance.sender:
        if instance.platform == 'Email':
            email = instance.sender.id
        elif instance.platform == 'WhatsApp':
            phone = instance.sender.id

    try:
        with transaction.atomic():
            # Create the request object using the instance directly
            request = Request.objects.create(
                message=instance,
                account=instance.sender,
                channel=instance.channel,
                email=email,
                phone=phone,
                display_text=instance.text,
                status='PENDING',
                metadata={
                    'created_from': 'message_signal',
                    'message_platform': instance.platform,
                    'creation_time': timezone.now().isoformat(),
                }
            )

            # Try to identify member
            request.identify_member()
            
            # Always proceed to categorization regardless of member identification
            request.categorize()

    except Exception as e:
        # Log the error but don't re-raise to avoid affecting message creation
        error_msg = f"Error creating request from message {instance.id}: {str(e)}"
        print(error_msg)
        logger = logging.getLogger(__name__)
        logger.error(error_msg, exc_info=True)

@receiver(post_save, sender=Request)
def mark_email_seen_on_request_completed(sender, instance, **kwargs):
    """
    When a Request is marked as COMPLETED, and the related Message is an email, and the channel's MARK_SEEN_WHEN is 'on_request_completed',
    connect to IMAP and mark the message as seen using the Message ID (IMAP UID).
    """
    if instance.status != 'COMPLETED':
        return
    msg = instance.message
    channel = instance.channel
    if not (msg and channel and msg.platform == 'Email'):
        return
    mark_seen_when = channel.config.get('MARK_SEEN_WHEN', 'never').lower()
    if mark_seen_when != 'on_request_completed':
        return
    # Get IMAP config
    imap_conf = channel.config.get('IMAP', {})
    host = imap_conf.get('host')
    port = imap_conf.get('port')
    use_ssl = imap_conf.get('use_ssl')
    email_address = channel.config.get('EMAIL_ADDRESS')
    password = channel.config.get('EMAIL_PASSWORD')
    if not all([host, port, email_address, password]):
        return
    # The Message ID is the IMAP UID
    uid = msg.id
    try:
        with IMAPClient(host, port=port, ssl=use_ssl) as server:
            server.login(email_address, password)
            server.select_folder('INBOX')
            # Mark as seen
            server.add_flags(uid, [SEEN])
            logger = logging.getLogger(__name__)
            logger.info(f"Marked email as seen for Message ID/UID {uid} on channel {channel.pk}")
    except Exception as e:
        logger = logging.getLogger(__name__)
        logger.warning(f"Failed to mark email as seen for Message ID/UID {uid} on channel {channel.pk}: {e}")


from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.utils import timezone
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_GET
from django.core.cache import cache
from django.core.exceptions import ValidationError
from unicom.models import Message, Channel
import uuid
from urllib.parse import urlparse, parse_qs, urlencode, urlunparse

def validate_tracking_id(tracking_id):
    """Validate the tracking ID format"""
    try:
        return str(uuid.UUID(str(tracking_id)))
    except (ValueError, AttributeError, TypeError):
        return None

def get_rate_limit_key(view_name, tracking_id, client_ip):
    """Generate a rate limit cache key"""
    return f"email_tracking:{view_name}:{tracking_id}:{client_ip}"

def check_rate_limit(key, max_requests=60, window=60):
    """
    Rate limiting function
    max_requests: maximum number of requests allowed in the time window
    window: time window in seconds
    """
    current = cache.get(key, 0)
    if current >= max_requests:
        return False
    
    cache.get_or_set(key, 0, window)
    cache.incr(key)
    return True

@csrf_exempt
@require_GET
def tracking_pixel(request, tracking_id):
    """
    Handle email open tracking via a 1x1 transparent pixel.
    Includes rate limiting and input validation.
    """
    # Validate tracking ID
    valid_id = validate_tracking_id(tracking_id)
    if not valid_id:
        return HttpResponse('Invalid tracking ID', status=400)

    # Rate limiting
    client_ip = request.META.get('HTTP_X_FORWARDED_FOR', request.META.get('REMOTE_ADDR'))
    rate_key = get_rate_limit_key('pixel', valid_id, client_ip)
    if not check_rate_limit(rate_key):
        return HttpResponse('Too many requests', status=429)

    # Cache the message lookup
    cache_key = f'tracking_message:{valid_id}'
    message = cache.get(cache_key)
    if not message:
        message = get_object_or_404(Message, tracking_id=valid_id)
        cache.set(cache_key, message, timeout=300)  # Cache for 5 minutes

    if not message.opened:
        message.opened = True
        message.time_opened = timezone.now()
        message.save(update_fields=['opened', 'time_opened'])
    
    # Return a 1x1 transparent GIF
    transparent_pixel = b'\x47\x49\x46\x38\x39\x61\x01\x00\x01\x00\x80\x00\x00\xff\xff\xff\x00\x00\x00\x21\xf9\x04\x01\x00\x00\x00\x00\x2c\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02\x44\x01\x00\x3b'
    return HttpResponse(transparent_pixel, content_type='image/gif')

@csrf_exempt
@require_GET
def link_click(request, tracking_id, link_index):
    """
    Handle email link click tracking and redirect to the original URL.
    Includes rate limiting, input validation, and efficient lookups.
    """
    # Validate tracking ID
    valid_id = validate_tracking_id(tracking_id)
    if not valid_id:
        return HttpResponse('Invalid tracking ID', status=400)

    # Validate link index
    try:
        link_index = int(link_index)
        if link_index < 0:
            raise ValueError
    except (ValueError, TypeError):
        return HttpResponse('Invalid link index', status=400)

    # Rate limiting
    client_ip = request.META.get('HTTP_X_FORWARDED_FOR', request.META.get('REMOTE_ADDR'))
    rate_key = get_rate_limit_key('link', valid_id, client_ip)
    if not check_rate_limit(rate_key):
        return HttpResponse('Too many requests', status=429)

    # Cache the message lookup
    cache_key = f'tracking_message:{valid_id}'
    message = cache.get(cache_key)
    if not message:
        message = get_object_or_404(Message, tracking_id=valid_id)
        cache.set(cache_key, message, timeout=300)  # Cache for 5 minutes

    try:
        original_url = message.raw.get('original_urls', [])[link_index]
        if not original_url:
            raise IndexError
    except (IndexError, KeyError):
        return HttpResponse('Invalid link', status=400)
    
    # Update link click tracking
    now = timezone.now()
    if not message.link_clicked:
        message.link_clicked = True
        message.time_link_clicked = now
    
    if original_url not in message.clicked_links:
        message.clicked_links.append(original_url)
    
    message.save(update_fields=['link_clicked', 'time_link_clicked', 'clicked_links'])
    
    # Redirect to the original URL with tracking id as a parameter if configured
    tracking_param = message.channel.config.get('TRACKING_PARAMETER_ID')
    if tracking_param:
        parsed = urlparse(original_url)
        query = parse_qs(parsed.query)
        query[tracking_param] = [str(tracking_id)]
        new_query = urlencode(query, doseq=True)
        new_url = urlunparse(parsed._replace(query=new_query))
        return HttpResponseRedirect(new_url)
    else:
        return HttpResponseRedirect(original_url) 


import pytest
import threading
from django.db import transaction

class InlineThread:
    """
    A fake Thread that just invokes its target immediately.
    """
    def __init__(self, target, args=(), kwargs=None, **_ignored):
        self._target = target
        self._args = args
        self._kwargs = kwargs or {}

    def start(self):
        # run in-band instead of spawning a new thread
        self._target(*self._args, **self._kwargs)

@pytest.fixture(autouse=True)
def run_validations_inline(monkeypatch):
    """
    Everywhere in tests, swap out:
      - Django’s transaction.on_commit => call immediately
      - threading.Thread       => InlineThread
    This makes your Channel.validate() fire synchronously
    on save(), so there are no background threads holding DB connections.
    """
    # 1) Make on_commit callbacks run immediately
    monkeypatch.setattr(
        transaction,
        "on_commit",
        lambda func, using=None: func()
    )

    # 2) Replace Thread with an inline runner
    monkeypatch.setattr(
        threading,
        "Thread",
        InlineThread
    )



from __future__ import annotations
from typing import TYPE_CHECKING
from django.db import models
from django.contrib.auth.models import User
from unicom.models.constants import channels
from django.contrib.postgres.fields import ArrayField
from django.core.validators import validate_email
from fa2svg.converter import revert_to_original_fa
import uuid
import re
from bs4 import BeautifulSoup
import base64
from .fields import DedupFileField, only_delete_file_if_unused

if TYPE_CHECKING:
    from unicom.models import Channel


class Message(models.Model):
    TYPE_CHOICES = [
        ('text', 'Text'),
        ('html', 'HTML'),
        ('image', 'Image'),
        ('audio', 'Audio'),
    ]
    id = models.CharField(max_length=500, primary_key=True)
    channel = models.ForeignKey('unicom.Channel', on_delete=models.CASCADE)
    platform = models.CharField(max_length=100, choices=channels)
    sender = models.ForeignKey('unicom.Account', on_delete=models.RESTRICT)
    user = models.ForeignKey(User, on_delete=models.RESTRICT, null=True, blank=True)
    chat = models.ForeignKey('unicom.Chat', on_delete=models.CASCADE, related_name='messages')
    is_outgoing = models.BooleanField(null=True, default=None, help_text="True for outgoing messages, False for incoming, None for internal")
    sender_name = models.CharField(max_length=100)
    subject = models.CharField(max_length=512, blank=True, null=True, help_text="Subject of the message (only for email messages)")
    text = models.TextField()
    html = models.TextField(
        blank=True, null=True,
        help_text="Full HTML body (only for email messages)"
    )
    to  = ArrayField(
        base_field=models.EmailField(validators=[validate_email]),
        blank=True,
        default=list,
        help_text="List of To: addresses",
    )
    cc  = ArrayField(
        base_field=models.EmailField(validators=[validate_email]),
        blank=True,
        default=list,
        help_text="List of Cc: addresses",
    )
    bcc = ArrayField(
        base_field=models.EmailField(validators=[validate_email]),
        blank=True,
        default=list,
        help_text="List of Bcc: addresses",
    )
    media = models.FileField(upload_to='media/', blank=True, null=True)
    reply_to_message = models.ForeignKey(
        'self', on_delete=models.SET_NULL, null=True, blank=True, related_name='replies')
    timestamp = models.DateTimeField()
    time_sent = models.DateTimeField(null=True, blank=True)
    time_delivered = models.DateTimeField(null=True, blank=True)
    time_seen = models.DateTimeField(null=True, blank=True)
    sent = models.BooleanField(default=False)
    delivered = models.BooleanField(default=False)
    seen = models.BooleanField(default=False)
    raw = models.JSONField()
    media_type = models.CharField(
        max_length=10,
        choices=TYPE_CHOICES,
        default='text'
    )
    # Email tracking fields
    tracking_id = models.UUIDField(default=uuid.uuid4, null=True, blank=True, help_text="Unique ID for tracking email opens and clicks")
    time_opened = models.DateTimeField(null=True, blank=True, help_text="When the email was first opened")
    opened = models.BooleanField(default=False, help_text="Whether the email has been opened")
    time_link_clicked = models.DateTimeField(null=True, blank=True, help_text="When a link in the email was first clicked")
    link_clicked = models.BooleanField(default=False, help_text="Whether any link in the email has been clicked")
    clicked_links = ArrayField(
        base_field=models.URLField(),
        blank=True,
        null=True,
        default=list,
        help_text="List of links that have been clicked"
    )

    def reply_with(self, msg_dict:dict) -> Message:
        """
        Reply to this message with a dictionary containing the response.
        The dictionary can contain 'text', 'html', 'file_path', etc.
        """
        from unicom.services.crossplatform.reply_to_message import reply_to_message
        return reply_to_message(self.channel, self, msg_dict)

    @property
    def original_content(self):
        return revert_to_original_fa(self.html) if self.platform == 'Email' else self.text

    @property
    def html_with_base64_images(self):
        """
        Returns the HTML content with all inline image shortlinks replaced by their original base64 data, if available.
        """
        if not self.html:
            return self.html
        soup = BeautifulSoup(self.html, 'html.parser')
        # Map shortlink src to base64 for all inline images
        images = {img.get_short_id(): img for img in getattr(self, 'inline_images', [])}
        for img_tag in soup.find_all('img'):
            src = img_tag.get('src', '')
            # Extract short id from src (e.g., /i/abc123 or full URL)
            m = re.search(r'/i/([A-Za-z0-9]+)', src)
            if m:
                short_id = m.group(1)
                image_obj = images.get(short_id)
                if image_obj:
                    # Read file and encode as base64
                    data = image_obj.file.read()
                    image_obj.file.seek(0)
                    mime = 'image/png'  # Default
                    if hasattr(image_obj.file, 'file') and hasattr(image_obj.file.file, 'content_type'):
                        mime = image_obj.file.file.content_type
                    elif image_obj.file.name:
                        import mimetypes
                        mime = mimetypes.guess_type(image_obj.file.name)[0] or 'image/png'
                    b64 = base64.b64encode(data).decode('ascii')
                    img_tag['src'] = f'data:{mime};base64,{b64}'
        return str(soup)

    class Meta:
        ordering = ['-timestamp']

    def __str__(self) -> str:
        return f"{self.platform}:{self.chat.name}->{self.sender_name}: {self.text}"


class EmailInlineImage(models.Model):
    file = DedupFileField(upload_to='email_inline_images/', hash_field='hash')
    email_message = models.ForeignKey(Message, on_delete=models.CASCADE, related_name='inline_images', null=True, blank=True)
    created_at = models.DateTimeField(auto_now_add=True)
    content_id = models.CharField(max_length=255, blank=True, null=True, help_text='Content-ID for cid: references in HTML')
    hash = models.CharField(max_length=64, blank=True, null=True, db_index=True, help_text='SHA256 hash of file for deduplication')

    def delete(self, *args, **kwargs):
        only_delete_file_if_unused(self, 'file', 'hash')
        super().delete(*args, **kwargs)

    def get_short_id(self):
        # Use base62 encoding of PK for short URLs
        import string
        chars = string.digits + string.ascii_letters
        n = self.pk
        s = ''
        if n == 0:
            return chars[0]
        while n > 0:
            n, r = divmod(n, 62)
            s = chars[r] + s
        return s



from django.contrib import admin
from django.db import models
from django.utils.translation import gettext_lazy as _
from unicom.models import Message, Update, Chat, Account, AccountChat, Channel
from django.utils.html import format_html
from unicom.views.chat_history_view import chat_history_view
from unicom.views.compose_view import compose_view
from django.urls import path, reverse
from django_ace import AceWidget
from django.contrib.admin import SimpleListFilter
from django.utils import timezone
from datetime import timedelta
from unicom.models import (
    Member,
    MemberGroup,
    RequestCategory,
    Request,
    MessageTemplate,
    DraftMessage,
    EmailInlineImage
)
from django import forms
from django.conf import settings
from django.utils.safestring import mark_safe
from unicom.services.get_public_origin import get_public_origin
from unicom.models.message_template import MessageTemplateInlineImage


class ArchiveStatusFilter(SimpleListFilter):
    title = _('Archive Status')
    parameter_name = 'archive_status'
    default_value = 'unarchived'

    def lookups(self, request, model_admin):
        return (
            ('unarchived', _('Unarchived')),
            ('archived', _('Archived')),
            ('all', _('All Chats')),
        )

    def queryset(self, request, queryset):
        value = self.value() or self.default_value
        if value == 'unarchived':
            return queryset.filter(is_archived=False)
        if value == 'archived':
            return queryset.filter(is_archived=True)
        # if value == 'all':
        return queryset

    def choices(self, changelist):
        value = self.value() or self.default_value
        for lookup, title in self.lookup_choices:
            yield {
                'selected': value == str(lookup),
                'query_string': changelist.get_query_string({self.parameter_name: lookup}),
                'display': title,
            }


class LastMessageTypeFilter(SimpleListFilter):
    title = _('Last Message Type')
    parameter_name = 'last_message_type'

    def lookups(self, request, model_admin):
        return (
            ('incoming', _('Needs Response')),
            ('outgoing', _('We Responded Last')),
            ('none', _('No Messages')),
        )

    def queryset(self, request, queryset):
        if self.value() == 'incoming':
            return queryset.filter(last_message__is_outgoing=False)
        if self.value() == 'outgoing':
            return queryset.filter(last_message__is_outgoing=True)
        if self.value() == 'none':
            return queryset.filter(last_message__isnull=True)


class LastMessageTimeFilter(SimpleListFilter):
    title = _('Last Activity')
    parameter_name = 'last_activity'

    def lookups(self, request, model_admin):
        return (
            ('1h', _('Within 1 hour')),
            ('24h', _('Within 24 hours')),
            ('7d', _('Within 7 days')),
            ('30d', _('Within 30 days')),
            ('old', _('Older than 30 days')),
            ('none', _('No activity')),
        )

    def queryset(self, request, queryset):
        now = timezone.now()
        if self.value() == '1h':
            return queryset.filter(last_message__timestamp__gte=now - timedelta(hours=1))
        if self.value() == '24h':
            return queryset.filter(last_message__timestamp__gte=now - timedelta(days=1))
        if self.value() == '7d':
            return queryset.filter(last_message__timestamp__gte=now - timedelta(days=7))
        if self.value() == '30d':
            return queryset.filter(last_message__timestamp__gte=now - timedelta(days=30))
        if self.value() == 'old':
            return queryset.filter(last_message__timestamp__lt=now - timedelta(days=30))
        if self.value() == 'none':
            return queryset.filter(last_message__isnull=True)


class MessageHistoryFilter(SimpleListFilter):
    title = _('Message History')
    parameter_name = 'message_history'

    def lookups(self, request, model_admin):
        return (
            ('has_both', _('Has both incoming & outgoing')),
            ('only_incoming', _('Only incoming messages')),
            ('only_outgoing', _('Only outgoing messages')),
            ('empty', _('No messages')),
        )

    def queryset(self, request, queryset):
        if self.value() == 'has_both':
            return queryset.filter(
                first_incoming_message__isnull=False,
                first_outgoing_message__isnull=False
            )
        if self.value() == 'only_incoming':
            return queryset.filter(
                first_incoming_message__isnull=False,
                first_outgoing_message__isnull=True
            )
        if self.value() == 'only_outgoing':
            return queryset.filter(
                first_incoming_message__isnull=True,
                first_outgoing_message__isnull=False
            )
        if self.value() == 'empty':
            return queryset.filter(
                first_message__isnull=True
            )


class ChatAdmin(admin.ModelAdmin):
    list_filter = (
        LastMessageTypeFilter,
        LastMessageTimeFilter,
        MessageHistoryFilter,
        ArchiveStatusFilter,
        'platform',
        'is_private',
        'channel',
    )
    list_display = ('chat_info',)
    search_fields = ('id', 'name', 'messages__text', 'messages__sender__name')
    actions = ['archive_chats', 'unarchive_chats']

    class Media:
        css = {
            'all': (
                'admin/css/chat_list.css',
                'https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.7.2/css/all.min.css',
            )
        }

    def get_queryset(self, request):
        # The filtering logic is now handled by the ArchiveStatusFilter.
        return super().get_queryset(request).order_by('-last_message__timestamp')

    def archive_chats(self, request, queryset):
        updated = queryset.update(is_archived=True)
        self.message_user(request, f'{updated} chat(s) have been archived.')
    archive_chats.short_description = 'Archive selected chats'

    def unarchive_chats(self, request, queryset):
        updated = queryset.update(is_archived=False)
        self.message_user(request, f'{updated} chat(s) have been unarchived.')
    unarchive_chats.short_description = 'Unarchive selected chats'

    def chat_info(self, obj):
        name = obj.name or obj.id
        last_message = obj.last_message
        last_message_text = last_message.text[:100] + '...' if last_message and last_message.text and len(last_message.text) > 100 else (last_message.text if last_message else 'No messages')
        
        # Determine message status indicators
        has_incoming = obj.first_incoming_message is not None
        has_outgoing = obj.first_outgoing_message is not None
        is_last_incoming = last_message and last_message.is_outgoing is False
        
        # Create status icons HTML
        status_icons = format_html(
            '<span class="chat-status-icons">{}{}{}</span>',
            format_html('<i class="fas fa-inbox" title="Has incoming messages"></i>') if has_incoming else '',
            format_html('<i class="fas fa-paper-plane" title="Has outgoing messages"></i>') if has_outgoing else '',
            format_html('<i class="fas fa-archive" title="Archived"></i>') if obj.is_archived else ''
        )
        
        # Create last message icon
        last_message_icon = ''
        if last_message:
            if is_last_incoming:
                last_message_icon = format_html('<i class="fas fa-reply pending-response" title="Pending response"></i>')
            else:
                last_message_icon = format_html('<i class="fas fa-check" title="Last message was outgoing"></i>')
        
        return format_html('''
            <a href="{}" class="chat-info-container{}">
                <div class="chat-header">
                    <span class="chat-name">{}</span>
                    {}
                </div>
                <div class="chat-message">
                    <span class="message-status-icon">{}</span>
                    <span class="message-text">{}</span>
                </div>
                <div class="chat-footer">
                    <span class="chat-channel">{}</span>
                    <span class="chat-time">{}</span>
                </div>
            </a>
            <style>
                .chat-info-container {{
                    display: block;
                    padding: 10px;
                    text-decoration: none;
                    color: var(--body-fg);
                    border-radius: 4px;
                    transition: background-color 0.2s;
                }}
                .chat-info-container:hover {{
                    background-color: var(--darkened-bg);
                }}
                .chat-info-container.archived {{
                    opacity: 0.7;
                }}
                .chat-header {{
                    margin-bottom: 5px;
                    display: flex;
                    justify-content: space-between;
                    align-items: center;
                }}
                .chat-name {{
                    font-weight: bold;
                    font-size: 1.1em;
                    color: var(--link-fg);
                }}
                .chat-status-icons {{
                    display: flex;
                    gap: 8px;
                    font-size: 0.8em;
                    color: var(--body-quiet-color);
                }}
                .chat-status-icons i {{
                    opacity: 0.7;
                }}
                .chat-message {{
                    color: var(--body-fg);
                    margin-bottom: 5px;
                    font-size: 0.9em;
                    opacity: 0.9;
                    display: flex;
                    align-items: flex-start;
                    gap: 8px;
                }}
                .message-status-icon {{
                    flex-shrink: 0;
                    margin-top: 3px;
                }}
                .message-status-icon .pending-response {{
                    color: #e74c3c;
                }}
                .message-text {{
                    flex-grow: 1;
                }}
                .chat-footer {{
                    display: flex;
                    justify-content: space-between;
                    font-size: 0.8em;
                    color: var(--body-quiet-color);
                }}
                .chat-channel {{
                    background-color: var(--selected-row);
                    padding: 2px 6px;
                    border-radius: 3px;
                }}
                .chat-time {{
                    color: var(--body-quiet-color);
                }}
            </style>
        ''',
        self.url_for_chat(obj.id),
        ' archived' if obj.is_archived else '',
        name,
        status_icons,
        last_message_icon,
        last_message_text,
        obj.channel,
        last_message.timestamp.strftime('%Y-%m-%d %H:%M') if last_message else 'Never'
        )
    chat_info.short_description = 'Chats'
    chat_info.admin_order_field = '-last_message__timestamp'

    def url_for_chat(self, id):
        return f"{id}/messages/"

    def get_urls(self):
        urls = super().get_urls()
        custom_urls = [
            path('<path:chat_id>/messages/', self.admin_site.admin_view(chat_history_view), name='chat-detail'),
            path('compose/', self.admin_site.admin_view(compose_view), name='chat-compose'),
        ]
        return custom_urls + urls

    def changelist_view(self, request, extra_context=None):
        extra_context = extra_context or {}
        extra_context['show_add_button'] = False  # Hide the default "Add" button
        return super().changelist_view(request, extra_context=extra_context)

    def has_add_permission(self, request):
        return False  # Disable the default add form

    def get_changelist_template(self, request):
        return "admin/unicom/chat/change_list.html"


class AccountChatAdmin(admin.ModelAdmin):
    list_filter = ('account__platform', )
    search_fields = ('account__name', 'chat__name') 


class AccountAdmin(admin.ModelAdmin):
    list_filter = ('platform', )
    search_fields = ('name', )

class ChannelAdmin(admin.ModelAdmin):
    list_filter = ('platform', )
    search_fields = ('name', )
    list_display = ('id', 'name', 'platform', 'active', 'confirmed_webhook_url', 'error')
    
    formfield_overrides = {
        models.JSONField: {'widget': AceWidget(mode='json', theme='twilight', width="100%", height="300px")},
    }

    class Media:
        js = ('unicom/js/channel_config.js',)

    def get_readonly_fields(self, request, obj=None):
        if obj:
            return ['active', 'confirmed_webhook_url', 'error']
        return super().get_readonly_fields(request, obj)

@admin.register(Member)
class MemberAdmin(admin.ModelAdmin):
    list_display = ('name', 'email', 'phone', 'group_list', 'created_at')
    list_filter = ('groups', 'created_at')
    search_fields = ('name', 'email', 'phone')
    readonly_fields = ('created_at', 'updated_at')
    filter_horizontal = ('allowed_categories',)
    
    def group_list(self, obj):
        return ", ".join([g.name for g in obj.groups.all()])
    group_list.short_description = "Groups"


@admin.register(MemberGroup)
class MemberGroupAdmin(admin.ModelAdmin):
    list_display = ('name', 'member_count', 'created_at')
    search_fields = ('name', 'description')
    readonly_fields = ('created_at', 'updated_at')
    filter_horizontal = ('members',)

    def member_count(self, obj):
        return obj.members.count()
    member_count.short_description = "Number of Members"


@admin.register(RequestCategory)
class RequestCategoryAdmin(admin.ModelAdmin):
    list_display = ('name', 'parent', 'sequence', 'is_active', 'is_public')
    list_filter = ('is_active', 'is_public', 'parent')
    search_fields = ('name', 'description')
    readonly_fields = ('created_at', 'updated_at')
    filter_horizontal = ('allowed_channels', 'authorized_members', 'authorized_groups')
    
    def formfield_for_dbfield(self, db_field, **kwargs):
        if db_field.name == 'processing_function':
            kwargs['widget'] = AceWidget(
                mode='python',
                theme='twilight',
                width="100%",
                height="300px"
            )
        return super().formfield_for_dbfield(db_field, **kwargs)
    
    def get_form(self, request, obj=None, **kwargs):
        form = super().get_form(request, obj, **kwargs)
        # Prevent a category from being its own parent
        if obj:
            form.base_fields['parent'].queryset = RequestCategory.objects.exclude(pk=obj.pk)
        
        # Set template code as initial value for new categories
        if not obj and 'processing_function' in form.base_fields:
            form.base_fields['processing_function'].initial = obj.get_template_code() if obj else RequestCategory().get_template_code()
        
        return form

    class Media:
        css = {
            'all': ('admin/css/forms.css',)
        }
        js = ('admin/js/jquery.init.js', 'admin/js/core.js',)


@admin.register(Request)
class RequestAdmin(admin.ModelAdmin):
    list_display = ('__str__', 'status', 'member_link', 'category', 'channel', 'created_at')
    list_display_links = ('__str__',)
    list_filter = (
        'status',
        'channel',
        'category',
        ('member', admin.RelatedOnlyFieldListFilter),
        ('created_at', admin.DateFieldListFilter),
    )
    search_fields = (
        'display_text',
        'message__text',
        'email',
        'phone',
        'member__name',
        'member__email',
        'member__phone',
        'metadata',
    )
    readonly_fields = (
        'created_at',
        'updated_at',
        'pending_at',
        'identifying_at',
        'categorizing_at',
        'queued_at',
        'processing_at',
        'completed_at',
        'failed_at',
        'error',
    )
    raw_id_fields = ('message', 'account', 'member', 'category')
    date_hierarchy = 'created_at'

    def member_link(self, obj):
        if obj.member:
            url = f"/admin/unicom/member/{obj.member.id}/change/"
            return format_html('<a href="{}">{}</a>', url, obj.member.name)
        return "-"
    member_link.short_description = "Member"

    fieldsets = (
        ('Message', {
            'fields': ('message', 'display_text')
        }),
        ('Basic Information', {
            'fields': ('status', 'error', 'account', 'channel', 'member')
        }),
        ('Contact Information', {
            'fields': ('email', 'phone')
        }),
        ('Categorization', {
            'fields': ('category',)
        }),
        ('Metadata', {
            'fields': ('metadata',),
            'classes': ('collapse',)
        }),
        ('Timestamps', {
            'fields': (
                'created_at',
                'updated_at',
                'pending_at',
                'identifying_at',
                'categorizing_at',
                'queued_at',
                'processing_at',
                'completed_at',
                'failed_at',
            ),
            'classes': ('collapse',)
        }),
    )

@admin.register(MessageTemplate)
class MessageTemplateAdmin(admin.ModelAdmin):
    list_display = ('title', 'category', 'created_at', 'updated_at')
    list_filter = ('category', 'channels')
    search_fields = ('title', 'description', 'content')
    readonly_fields = ('created_at', 'updated_at')
    filter_horizontal = ('channels',)
    
    fieldsets = (
        (None, {
            'fields': ('title', 'category')
        }),
        (_('Template Content'), {
            'fields': ('description', 'content'),
            'classes': ('tinymce-content',),
        }),
        (_('Availability'), {
            'fields': ('channels',),
        }),
        (_('Metadata'), {
            'fields': ('created_at', 'updated_at'),
            'classes': ('collapse',),
        }),
    )

    def get_form(self, request, obj=None, **kwargs):
        form = super().get_form(request, obj, **kwargs)
        # Only include the local tinymce_init.js, not the CDN script
        form.Media = type('Media', (), {
            'css': {'all': ('admin/css/forms.css',)},
            'js': (
                'unicom/js/tinymce_init.js',
            )
        })
        return form

    def render_change_form(self, request, context, *args, **kwargs):
        context['tinymce_api_key'] = settings.UNICOM_TINYMCE_API_KEY
        return super().render_change_form(request, context, *args, **kwargs)

    def formfield_for_dbfield(self, db_field, **kwargs):
        if db_field.name == 'content':
            kwargs['widget'] = forms.Textarea(attrs={
                'class': 'tinymce',
                'data-tinymce': 'true'
            })
        return super().formfield_for_dbfield(db_field, **kwargs)

class DraftScheduleFilter(SimpleListFilter):
    title = _('Schedule Status')
    parameter_name = 'schedule_status'
    default_value = 'pending'

    def lookups(self, request, model_admin):
        """
        Returns a list of tuples. The first element in each
        tuple is the coded value for the option that will appear in the URL query.
        The second element is the human-readable name for the option that will
        appear in the right sidebar.
        """
        return (
            ('pending', _('Pending Approval')),
            ('all', _('All')),
            ('scheduled', _('Scheduled & Approved')),
            ('past_due', _('Past Due')),
            ('draft', _('Draft')),
        )

    def queryset(self, request, queryset):
        """
        Returns the filtered queryset based on the value
        provided in the query string and retrievable via `self.value()`.
        """
        value = self.value() or self.default_value
        now = timezone.now()

        if value == 'pending':
            return queryset.filter(status='scheduled', is_approved=False, send_at__gt=now)
        if value == 'scheduled':
            return queryset.filter(status='scheduled', is_approved=True, send_at__gt=now)
        if value == 'past_due':
            return queryset.filter(status='scheduled', send_at__lt=now)
        if value == 'draft':
            return queryset.filter(status='draft')
        if value == 'all':
            return queryset
        return queryset

    def choices(self, changelist):
        """
        Override the default choices to prevent the automatic "All" link and
        to select our custom default.
        """
        value = self.value() or self.default_value
        for lookup, title in self.lookup_choices:
            yield {
                'selected': value == str(lookup),
                'query_string': changelist.get_query_string({self.parameter_name: lookup}),
                'display': title,
            }


@admin.register(DraftMessage)
class DraftMessageAdmin(admin.ModelAdmin):
    list_display = ('message_preview',)
    list_filter = (
        DraftScheduleFilter,
        'status',
        'is_approved',
        'channel',
        'created_by',
    )
    search_fields = ('text', 'html', 'subject', 'to', 'cc', 'bcc', 'chat_id')
    readonly_fields = ('created_at', 'updated_at', 'sent_at', 'error_message', 'sent_message')
    actions = ['approve_drafts', 'unapprove_drafts']

    fieldsets = (
        (None, {
            'fields': ('channel',)
        }),
        (_('Message Content'), {
            'fields': ('text', 'html'),
            'classes': ('tinymce-content',),
        }),
        (_('Email Specific'), {
            'fields': ('to', 'cc', 'bcc', 'subject'),
            'classes': ('collapse',),
        }),
        (_('Chat Specific'), {
            'fields': ('chat_id',),
            'classes': ('collapse',),
        }),
        (_('Scheduling & Approval'), {
            'fields': ('send_at', 'is_approved', 'status'),
        }),
        (_('Metadata'), {
            'fields': ('created_by', 'created_at', 'updated_at', 'sent_at', 'sent_message', 'error_message'),
            'classes': ('collapse',),
        }),
    )

    class Media:
        css = {
            'all': (
                'admin/css/chat_list.css',
                'https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.7.2/css/all.min.css',
            )
        }

    def save_model(self, request, obj, form, change):
        if not change:  # If this is a new object
            obj.created_by = request.user
        super().save_model(request, obj, form, change)
    
    def get_form(self, request, obj=None, **kwargs):
        form = super().get_form(request, obj, **kwargs)
        # Add TinyMCE for HTML content
        form.Media = type('Media', (), {
            'css': {'all': ('admin/css/forms.css',)},
            'js': (
                'unicom/js/tinymce_init.js',
            )
        })
        return form
    
    def formfield_for_dbfield(self, db_field, **kwargs):
        if db_field.name == 'html':
            kwargs['widget'] = forms.Textarea(attrs={
                'class': 'tinymce',
                'data-tinymce': 'true'
            })
        return super().formfield_for_dbfield(db_field, **kwargs)

    def get_queryset(self, request):
        return super().get_queryset(request)

    def approve_drafts(self, request, queryset):
        updated = queryset.update(is_approved=True)
        self.message_user(request, f'{updated} draft(s) have been approved.')
    approve_drafts.short_description = 'Approve selected drafts'

    def unapprove_drafts(self, request, queryset):
        updated = queryset.update(is_approved=False)
        self.message_user(request, f'{updated} draft(s) have been unapproved.')
    unapprove_drafts.short_description = 'Unapprove selected drafts'

    def message_preview(self, obj):
        # Determine message type and status indicators
        is_email = obj.channel.platform == 'Email'
        is_pending = obj.status == 'scheduled' and not obj.is_approved
        is_past_due = obj.status == 'scheduled' and obj.send_at and obj.send_at < timezone.now()
        
        # Create status icons
        status_icons = format_html(
            '<span class="draft-status-icons">{}{}{}{}</span>',
            format_html('<i class="fas fa-clock text-warning" title="Pending Approval"></i>') if is_pending else '',
            format_html('<i class="fas fa-exclamation-circle text-danger" title="Past Due"></i>') if is_past_due else '',
            format_html('<i class="fas fa-check-circle text-success" title="Approved"></i>') if obj.is_approved else '',
            format_html('<i class="fas fa-envelope" title="Email"></i>') if is_email else format_html('<i class="fas fa-comment" title="Chat"></i>')
        )

        # Prepare content preview
        if is_email and obj.html:
            # Use an iframe for email preview with auto-resize
            iframe_id = f"email-preview-iframe-{obj.pk}"
            content_preview = format_html(
                '''
                <iframe id="{}" style="background-color: white; width:100%;border:none;overflow:hidden;min-height:40px;" scrolling="no" frameborder="0" allowtransparency="true"></iframe>
                <script type="text/javascript">
                (function() {{
                    var iframe = document.getElementById('{}');
                    if (!iframe) return;
                    var doc = iframe.contentDocument || iframe.contentWindow.document;
                    var html = {};
                    doc.open();
                    doc.write(html);
                    doc.close();
                    function resizeIframe() {{
                        setTimeout(function() {{
                            if (!iframe.contentWindow.document.body) return;
                            var body = iframe.contentWindow.document.body;
                            var html = iframe.contentWindow.document.documentElement;
                            var height = Math.max(
                                body.scrollHeight,
                                body.offsetHeight,
                                html.clientHeight,
                                html.scrollHeight,
                                html.offsetHeight
                            );
                            iframe.style.height = height + 'px';
                        }}, 50);
                    }}
                    // Resize on load and after images load
                    iframe.onload = resizeIframe;
                    doc.addEventListener('DOMContentLoaded', resizeIframe);
                    var imgs = doc.images;
                    for (var i = 0; i < imgs.length; i++) {{
                        imgs[i].addEventListener('load', resizeIframe);
                    }}
                    // Fallback resize after a short delay
                    setTimeout(resizeIframe, 500);
                }})();
                </script>
                ''',
                iframe_id,
                iframe_id,
                mark_safe(repr(obj.html))
            )
        else:
            content_preview = obj.text if obj.text else 'No content'

        # Format recipients for email
        recipients = ''
        if is_email:
            to_list = ', '.join(obj.to) if obj.to else ''
            cc_list = f' (cc: {", ".join(obj.cc)})' if obj.cc else ''
            recipients = f'{to_list}{cc_list}' if to_list or cc_list else 'No recipients'

        return format_html('''
            <div class="draft-message-container">
                <div class="draft-header">
                    <div class="draft-title">
                        <span class="draft-subject">{}</span>
                        {}
                    </div>
                    <div class="draft-meta">
                        <span class="draft-channel">{}</span>
                        <span class="draft-time" title="Send At">{}</span>
                    </div>
                </div>
                {}
                <div class="draft-content">
                    {}
                </div>
                <div class="draft-footer">
                    <span class="draft-creator">By: {}</span>
                    <span class="draft-created">Created: {}</span>
                </div>
            </div>
            <style>
                .draft-message-container {{
                    padding: 15px;
                    border-radius: 4px;
                    background: var(--body-bg);
                    margin: 5px 0;
                }}
                .draft-header {{
                    display: flex;
                    justify-content: space-between;
                    align-items: flex-start;
                    margin-bottom: 10px;
                }}
                .draft-title {{
                    display: flex;
                    align-items: center;
                    gap: 10px;
                }}
                .draft-subject {{
                    font-weight: bold;
                    font-size: 1.1em;
                    color: var(--link-fg);
                }}
                .draft-status-icons {{
                    display: flex;
                    gap: 8px;
                }}
                .draft-status-icons i {{
                    font-size: 1.1em;
                }}
                .text-warning {{
                    color: #f39c12;
                }}
                .text-danger {{
                    color: #e74c3c;
                }}
                .text-success {{
                    color: #2ecc71;
                }}
                .draft-meta {{
                    display: flex;
                    gap: 15px;
                    align-items: center;
                }}
                .draft-channel {{
                    background-color: var(--selected-row);
                    padding: 2px 6px;
                    border-radius: 3px;
                    font-size: 0.9em;
                }}
                .draft-time {{
                    color: var(--body-quiet-color);
                    font-size: 0.9em;
                }}
                .draft-recipients {{
                    font-size: 0.9em;
                    color: var(--body-quiet-color);
                    margin: 5px 0;
                }}
                .draft-content {{
                    margin: 10px 0;
                    padding: 10px;
                    background: var(--darkened-bg);
                    border-radius: 4px;
                    max-height: 300px;
                    overflow-y: auto;
                }}
                .draft-footer {{
                    display: flex;
                    justify-content: space-between;
                    font-size: 0.8em;
                    color: var(--body-quiet-color);
                    margin-top: 10px;
                }}
            </style>
        ''',
        obj.subject or 'No Subject',
        status_icons,
        obj.channel,
        obj.send_at.strftime('%Y-%m-%d %H:%M') if obj.send_at else 'No schedule',
        format_html('<div class="draft-recipients">{}</div>', recipients) if recipients else '',
        content_preview,
        (obj.created_by.get_full_name() or obj.created_by.username) if obj.created_by else 'Unknown',
        obj.created_at.strftime('%Y-%m-%d %H:%M')
        )
    message_preview.short_description = 'Draft Messages'

class EmailInlineImageAdmin(admin.ModelAdmin):
    list_display = ('id', 'file', 'email_message', 'created_at', 'serving_link')
    readonly_fields = ('serving_link',)

    def serving_link(self, obj):
        if not obj.pk:
            return "(save to get link)"
        shortid = obj.get_short_id()
        path = reverse('inline_image', kwargs={'shortid': shortid})
        url = f"{get_public_origin().rstrip('/')}{path}"
        return format_html('<a href="{}" target="_blank">{}</a>', url, url)
    serving_link.short_description = "Serving Link"

class MessageTemplateInlineImageAdmin(admin.ModelAdmin):
    list_display = ('id', 'file', 'template', 'created_at', 'serving_link')
    readonly_fields = ('serving_link',)

    def serving_link(self, obj):
        if not obj.pk:
            return "(save to get link)"
        shortid = obj.get_short_id()
        path = reverse('template_inline_image', kwargs={'shortid': shortid})
        url = f"{get_public_origin().rstrip('/')}{path}"
        return format_html('<a href="{}" target="_blank">{}</a>', url, url)
    serving_link.short_description = "Serving Link"

admin.site.register(Channel, ChannelAdmin)
admin.site.register(Message)
admin.site.register(Update)
admin.site.register(Chat, ChatAdmin)
admin.site.register(Account, AccountAdmin)
admin.site.register(AccountChat, AccountChatAdmin)
admin.site.register(EmailInlineImage, EmailInlineImageAdmin)
admin.site.register(MessageTemplateInlineImage, MessageTemplateInlineImageAdmin)



# Generated by Django 4.2.20 on 2025-06-07 03:50

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('unicom', '0002_emailinlineimage'),
    ]

    operations = [
        migrations.AlterField(
            model_name='emailinlineimage',
            name='email_message',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='inline_images', to='unicom.message'),
        ),
    ]



from unicom.services.email.save_email_message import save_email_message
from imapclient import IMAPClient, SEEN
from imapclient.exceptions import IMAPClientError
from django.db import connections
import imaplib
import time
import logging

logger = logging.getLogger(__name__)


def listen_to_IMAP(channel):
    """
    Connects to the IMAP server defined in channel.config and listens via IDLE.
    Fetches new messages and hands them off to save_email_message().
    This runs indefinitely, with automatic reconnects on failure.
    """
    email_address = channel.config['EMAIL_ADDRESS']
    password      = channel.config['EMAIL_PASSWORD']
    imap_conf     = channel.config['IMAP']
    host          = imap_conf['host']
    port          = imap_conf['port']
    use_ssl       = imap_conf['use_ssl']

    logger.info(f"Channel {channel.pk}: Starting IMAP listener for {email_address} at {host}:{port} (SSL={use_ssl})")

    while True:
        try:
            with IMAPClient(host, port=port, ssl=use_ssl) as server:
                server.login(email_address, password)
                server.select_folder('INBOX')
                # caps = server.capabilities()
                mark_seen_on = channel.config.get('mark_seen_on', 'never')
                # Immediately fetch any older unseen messages on startup
                uids = server.search(['UNSEEN'])
                for uid in uids:
                    try:
                        resp = server.fetch(uid, ['BODY.PEEK[]'])
                        raw = resp[uid][b'BODY[]']
                        msg = save_email_message(channel, raw)
                        # logger.info(f"Channel {channel.pk}: Found email {msg.id} (uid={uid})")
                    except Exception as e:
                        logger.exception(f"Channel {channel.pk}: Failed to process UID {uid}: {e}")

                if mark_seen_on == 'on_save':
                    if uids:
                        server.add_flags(uids, [SEEN])
                        logger.info(f"Channel {channel.pk}: Marked {len(uids)} messages as SEEN on startup (on_save).")

                logger.info(f"Channel {channel.pk}: Connected to {host}:{port}, entering IDLE…")

                while True:
                    idle_tag = None
                    try:
                        idle_tag = server.idle()
                        responses = server.idle_check(timeout=300)
                    except (imaplib.IMAP4.abort, 
                            imaplib.IMAP4.error, 
                            IMAPClientError,
                            ConnectionResetError, 
                            OSError) as e:
                        if 'Unexpected IDLE response' in str(e) or 'Broken pipe' in str(e) or 'Connection reset by peer' in str(e):
                            break # ignore repeated IDLE errors for now TODO: prevent the error from occuring alltogether
                        logger.warning(f"Channel {channel.pk}: IMAP idle lost: {e}, reconnecting…")
                        break
                    finally:
                        if idle_tag:
                            try:
                                server.idle_done()
                                time.sleep(1)
                            except Exception as e:
                                logger.warning(f"Channel {channel.pk}: Failed to end IDLE: {e}")

                    if not responses:
                        continue

                    uids = server.search(['UNSEEN'])
                    for uid in uids:
                        try:
                            resp = server.fetch(uid, ['BODY.PEEK[]'])
                            raw = resp[uid][b'BODY[]']
                            msg = save_email_message(channel, raw)
                            logger.info(f"Channel {channel.pk}: Saved email {msg.id} (uid={uid})")
                            if mark_seen_on == 'on_save':
                                server.add_flags(uid, [SEEN])
                            logger.debug(f"Incoming email - Message-ID: {msg.id}, In-Reply-To: {msg.raw.get('In-Reply-To') if msg.raw else 'None'}")
                            logger.debug(f"Associated with chat: {msg.chat_id}")
                        except Exception:
                            logger.exception(f"Channel {channel.pk}: Failed to process UID {uid}")
                        finally:
                            connections.close_all()

        except Exception as e:
            logger.exception(f"Channel {channel.pk}: Fatal IMAP error: {e}, reconnecting in 30s…")
            time.sleep(3)
        finally:
            # Ensure we close all connections to avoid leaks
            connections.close_all()


from __future__ import annotations
from typing import TYPE_CHECKING
from unicom.services.telegram.send_telegram_message import send_telegram_message
from unicom.services.whatsapp.send_whatsapp_message import send_whatsapp_message
from unicom.services.email.send_email_message import send_email_message
from django.contrib.auth.models import User

if TYPE_CHECKING:
    from unicom.models import Channel, Message


def send_message(channel: Channel, msg:dict, user:User=None) -> Message:
    """
    The msg dict must include platform-specific required fields:

    For Email:
        New threads:
            - 'to': list of recipient email addresses
            - 'subject': required subject line
        Replies:
            - Either 'chat_id' or 'reply_to_message_id'
            - Subject is optional (derived from parent if not provided)
    
    For Telegram/WhatsApp:
        - 'chat_id' and 'text' are required
    """
    if channel.platform == 'Telegram':
        return send_telegram_message(channel, msg, user)
    elif channel.platform == 'WhatsApp':
        return send_whatsapp_message(channel, msg, user)
    elif channel.platform == 'Email':
        return send_email_message(channel, msg, user)


from setuptools import setup, find_packages

setup(
    name='django-unicom',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    description='Unified communication layer for Django (Telegram, WhatsApp, Email)',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Meena (Menas) Erian',
    author_email='hi@menas.pro',
    url='https://github.com/meena-erian/unicom',
    packages=find_packages(exclude=['tests*']),
    include_package_data=True,
    install_requires=[
        'Django>=3.2',
        'psycopg2-binary>=2.9,<3.0',
        'urllib3>=2.4.0,<3.0',
        'dnspython>=2.7.0,<3.0',
        'requests>=2.31.0,<3.0',
        'IMAPClient>=3.0.1',
        'dnspython>=2.7.0,<3.0',
        'Pillow>=10.4.0',
        'django-ace>=1.39.2',
        'fa2svg==0.1.6',
        'pytz>=2024.1',
        'openai',
    ],
    extras_require={
        'dev': [
            'pytest>=8.3.5,<9.0',
            'pytest-django>=4.11.1,<5.0',
            'charset-normalizer>=3.1.0,<4.0',
            'python-dotenv>=1.0,<2.0',
        ],
    },
    classifiers=[
        'Framework :: Django',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
    ],
)



from django.conf import settings
import base64
import uuid
import os



def decode_base64_image(base64_string, output_subdir="media"):
    """
    Decodes a base64-encoded image and writes it into:
      <settings.MEDIA_ROOT>/<output_subdir>/<uuid>.jpg
    Returns the *relative* path "media/<uuid>.jpg" (i.e. no leading slash).
    
    So physically:   /my_project/media/media/<uuid>.jpg
    DB storage:      "media/<uuid>.jpg"
    Final serve URL: <MEDIA_URL>/media/<uuid>.jpg => /media/media/<uuid>.jpg
    """
    # 1) Physical directory
    physical_dir = os.path.join(settings.MEDIA_ROOT, output_subdir)
    os.makedirs(physical_dir, exist_ok=True)

    # 2) Generate filename
    filename = f"{uuid.uuid4()}.jpg"
    full_path = os.path.join(physical_dir, filename)

    # 3) Write the file
    with open(full_path, "wb") as f:
        f.write(base64.b64decode(base64_string))

    # 4) Return the relative DB path: "media/<uuid>.jpg"
    return os.path.join(output_subdir, filename)


from bs4 import BeautifulSoup
from django.urls import reverse
from django.conf import settings
from unicom.services.get_public_origin import get_public_origin
import re
import uuid

# Unique identifiers for our tracking elements
TRACKING_PIXEL_CLASS = 'e-px-' # Will be followed by tracking ID
TRACKING_LINK_CLASS = 'e-lc-'  # Will be followed by tracking ID
TRACKING_LINK_INDEX_ATTR = 'data-link-index'


def add_tracking_pixel(html_content: str, tracking_id: uuid.UUID) -> str:
    """
    Add a 1x1 transparent tracking pixel to HTML email content.
    The pixel will have a unique class name that includes the tracking ID.
    """
    if not html_content:
        return html_content

    soup = BeautifulSoup(html_content, 'html.parser')
    
    # Create tracking pixel with unique identifier
    tracking_url = f"{get_public_origin()}{reverse('e_px', args=[tracking_id])}"
    pixel = soup.new_tag('img', 
                        src=tracking_url,
                        width="1",
                        height="1",
                        style="display:none",
                        **{'class': f"{TRACKING_PIXEL_CLASS}{tracking_id}"})
    
    # Add pixel to the end of the body, or create body if it doesn't exist
    if soup.body:
        soup.body.append(pixel)
    else:
        body = soup.new_tag('body')
        body.append(pixel)
        soup.append(body)
    
    return str(soup)


def wrap_links(html_content: str, tracking_id: uuid.UUID) -> tuple[str, list[str]]:
    """
    Wrap all links in the HTML content with tracking URLs.
    Each wrapped link will have a unique class and data attribute for the link index.
    Returns tuple of (modified_html, original_urls)
    """
    if not html_content:
        return html_content, []

    soup = BeautifulSoup(html_content, 'html.parser')
    original_urls = []
    
    # Find all links
    for link in soup.find_all('a', href=True):
        original_url = link['href']
        
        # Skip mailto: links and anchor links
        if original_url.startswith(('mailto:', '#', 'tel:')):
            continue
            
        # Store original URL and create tracking URL with index
        original_urls.append(original_url)
        link_index = len(original_urls) - 1
        tracking_url = f"{get_public_origin()}{reverse('e_lc', args=[tracking_id, link_index])}"
        link['href'] = tracking_url
        link['class'] = link.get('class', []) + [f"{TRACKING_LINK_CLASS}{tracking_id}"]
        link[TRACKING_LINK_INDEX_ATTR] = str(link_index)
    
    return str(soup), original_urls


def prepare_email_for_tracking(html_content: str, tracking_id: uuid.UUID) -> tuple[str, list[str]]:
    """
    Prepare an HTML email for tracking by adding a tracking pixel and wrapping links.
    Returns tuple of (modified_html, original_urls)
    """
    if not html_content:
        return html_content, []
        
    # First wrap all links
    html_with_wrapped_links, original_urls = wrap_links(html_content, tracking_id)
    
    # Then add tracking pixel
    final_html = add_tracking_pixel(html_with_wrapped_links, tracking_id)
    
    return final_html, original_urls


def remove_tracking(html_content: str, original_urls: list[str]) -> str:
    """
    Remove tracking elements and restore original URLs in an HTML email.
    """
    if not html_content:
        return html_content

    soup = BeautifulSoup(html_content, 'html.parser')
    
    # Remove tracking pixels
    for pixel in soup.find_all('img', class_=re.compile(f'^{TRACKING_PIXEL_CLASS}')):
        pixel.decompose()
    
    # Restore original URLs
    for link in soup.find_all('a', class_=re.compile(f'^{TRACKING_LINK_CLASS}')):
        try:
            link_index = int(link.get(TRACKING_LINK_INDEX_ATTR, -1))
            if 0 <= link_index < len(original_urls):
                link['href'] = original_urls[link_index]
            # Remove tracking attributes
            link['class'] = [c for c in link.get('class', []) 
                           if not c.startswith((TRACKING_LINK_CLASS, TRACKING_PIXEL_CLASS))]
            if not link['class']:  # Remove empty class attribute
                del link['class']
            del link[TRACKING_LINK_INDEX_ATTR]
        except (ValueError, IndexError):
            continue
    
    return str(soup) 


from unicom.models import Message, Chat, Account, AccountChat
from datetime import datetime
from django.contrib.auth.models import User

def save_internal_message(message_data: dict, user: User=None,):
    platform = 'Internal'
    sender_id = message_data.get('from')['id']
    sender_name = message_data.get('from')['first_name']
    is_outgoing = message_data.get('from')['is_bot']
    chat_id = message_data.get('chat')['id']
    chat_is_private = message_data.get('chat')["type"] == "private"
    chat_name = message_data.get('chat')["title"]
    message_id = message_data.get('message_id')
    media_type = message_data.get('media_type', 'text')  # or 'image', 'audio'
    media_file = message_data.get('media')  # might be a path, or None
    text = message_data.get('text') or message_data.get('caption')
    timestamp = datetime.fromtimestamp(message_data.get('date'))
    chat = Chat.objects.filter(platform=platform, id=chat_id)
    account = Account.objects.filter(platform=platform, id=sender_id)
    if not chat.exists():
        chat = Chat(platform=platform, id=chat_id, is_private=chat_is_private, name=chat_name)
        chat.save()
    else:
        chat = chat.get()

    if not account.exists():
        account = Account(
            platform=platform,
            id=sender_id,
            name=sender_name,
            is_bot=is_outgoing,
            raw=message_data.get('from')
        )
        account.save()
    else:
        account = account.get()

    account_chat = AccountChat.objects.filter(account=account, chat=chat)
    if not account_chat.exists():
        account_chat = AccountChat(account=account, chat=chat)
        account_chat.save()
    else:
        account_chat = account_chat.get()

    if message_data.get('reply_to_message'):
        reply_to_message_id = message_data.get('reply_to_message')['message_id']
        try:
            reply_to_message = Message.objects.get(platform=platform, chat_id=chat_id, id=reply_to_message_id)
        except Message.DoesNotExist:
            reply_to_message = None
    else:
        reply_to_message = None

    message, created = Message.objects.get_or_create(
        platform=platform,
        chat_id=chat_id,
        id=message_id,
        defaults={
            'sender': account,
            'channel': chat.channel,
            'sender_name': sender_name,
            'user': user,
            'text': text,
            'reply_to_message': reply_to_message,
            'timestamp': timestamp,
            'raw': message_data,
            'media_type': media_type,
            'media': media_file,
            'is_outgoing': is_outgoing,
        }
    )
    if not created:
        print("Duplicate message discarded")
    return message



import requests


def get_file_path(TelegramCredentials, file_id: str) -> str:
    TELEGRAM_API_TOKEN = TelegramCredentials["TELEGRAM_API_TOKEN"]
    url = f'https://api.telegram.org/bot{TELEGRAM_API_TOKEN}/getFile?file_id={file_id}'
    response = requests.get(url)
    response_json = response.json()
    file_path = response_json['result']['file_path']
    return file_path